/*---------------------------------------------------------------Benito7g.c
 The benito7g is an AT90USB162 based board produced by tempusdictum.
 This firmware turns the benito7 into a usb to serial converter with a
 2ms pulse instead of the DTR signal. This is designed to reset 
 microcontrollers with serial bootloaders. 
 -------------------------------------------------------------------------------
 Based on BenitoSerial.c CopyLeft 2008 Donald Delmar Davis (don@digithink.com)
  based on USBtoSerial.[c-h]
 MyUSB Library Copyright (C) Dean Camera, 2008.
  dean [at] fourwalledcubicle [dot] com www.fourwalledcubicle.com
 Released under the LGPL Licence, Version 3
-------------------------------------------------------------------------------*/
/*	WINDOWS USERS,
    (windows has a driver for this but it cant figure it out.) 
    Before running, you will need to install the INF file that
	is located in the source project directory. This will enable
	Windows to use its inbuilt CDC drivers, negating the need
	for special Windows drivers for the device. To install,
	right-click the .INF file and choose the Install option.
*/

/*
	USB Mode:           Device
	USB Class:          Communications Device Class (CDC)
	USB Subclass:       Abstract Control Model (ACM)
	Relevant Standards: USBIF CDC Class Standard
	Usable Speeds:      Full Speed Mode
*/

/*-----------------------------------------------------------------------------*
 ************************************ GLOBALS **********************************
 *-----------------------------------------------------------------------------*/	

#include "Benito7g.h"
/* Scheduler Task List */
TASK_LIST
{
	{ Task: USB_USBTask          , TaskStatus: TASK_STOP },
	{ Task: CDC_Task             , TaskStatus: TASK_STOP },
	{ Task: CDC_DTR_PULSE_Task	 , TaskStatus: TASK_STOP },
	{ Task: CDC_TXD_POV_Task	 , TaskStatus: TASK_STOP },
	{ Task: CDC_RXD_POV_Task	 , TaskStatus: TASK_STOP },	
};

/* Globals: */
CDC_Line_Coding_t LineCoding = { BaudRateBPS: 19200,
                                 CharFormat:  OneStopBit,
                                 ParityType:  Parity_None,
                                 DataBits:    8            };

RingBuff_t        Rx_Buffer;
RingBuff_t        Tx_Buffer;

volatile bool     Transmitting = false;
volatile bool     RXD_Flag = false;
volatile bool     TXD_Flag = false;
volatile bool     DTR_Flag = false;

/*-----------------------------------------------------------------------------*
 **************************** INITIALIZATION AND MAIN **************************
 *-----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------int main(void)
 Set up the timer,serial,iopins and start scheduler.
-------------------------------------------------------------------------------*/

int main(void)
{
	/* Disable watchdog if enabled by bootloader/fuses */
	MCUSR &= ~(1 << WDRF);
	wdt_disable();

	/* Disable Clock Division */
	SetSystemClockPrescaler(0);

    /* Millisecond timer initialization, with output compare interrupt enabled */
	OCR0A  = 0x7D;
	TCCR0A = (1 << WGM01);
	TCCR0B = ((1 << CS01) | (1 << CS00));
	TIMSK0 = (1 << OCIE0A);
	

	/* Hardware Initialization */
//	LEDS_DDR  |=  _BV(TXD_LED_PIN)|_BV(RXD_LED_PIN);
//	LEDS_PORT |=  _BV(TXD_LED_PIN)|_BV(RXD_LED_PIN);
//	DTR_PIN_DDR  |=  _BV(DTR_PIN);
//	DTR_PIN_PORT |= _BV(DTR_PIN); // PULL DTR HIGH
	//RTS_PIN_DDR |= _BV(RTS_PIN);
	//RTS_PIN_PORT |= _BV(RTS_PIN);
	ReconfigureUSART();
	
	
	/* Ringbuffer Initialization */
	Buffer_Initialize(&Rx_Buffer);
	Buffer_Initialize(&Tx_Buffer);
	
	/* Indicate USB not ready */
//	LEDS_PORT &= ~(_BV(TXD_LED_PIN)|_BV(RXD_LED_PIN));  
	
	/* Initialize Scheduler so that it can be used */
	Scheduler_Init();

	/* Initialize USB Subsystem */
	USB_Init();

	/* Scheduling - routine never returns, so put this last in the main function */
	Scheduler_Start();
}

/*----------------------------------------------ISR(TIMER0_COMPA_vect, ISR_BLOCK)
 * Increment scheduler tick counter once each millisecond 
 * Gleaned from MyUsb test demo
 *-----------------------------------------------------------------------------*/
 
ISR(TIMER0_COMPA_vect, ISR_BLOCK)
{
	/* Scheduler test - increment scheduler tick counter once each millisecond */
	Scheduler_TickCounter++;
}


/*-----------------------------------------------------------------------------*
 *********************************** USB SECTION *******************************
 *-----------------------------------------------------------------------------*/

/*-----------------------------------------------------EVENT_HANDLER(USB_Connect)
 * start USB_USBTask, update status
 *-----------------------------------------------------------------------------*/

EVENT_HANDLER(USB_Connect)
{
	/* Start USB management task */
	Scheduler_SetTaskMode(USB_USBTask, TASK_RUN);

	/* Indicate USB enumerating */
//	LEDS_PORT &= ~(_BV(RXD_LED_PIN));  //rxd on
//	LEDS_PORT |= _BV(TXD_LED_PIN);  //txd off
}

/*--------------------------------------------------EVENT_HANDLER(USB_Disconnect)
 *  stop CDC_Task and USB_USBTask, update status
 *-----------------------------------------------------------------------------*/

EVENT_HANDLER(USB_Disconnect)
{
	/* Stop running CDC and USB management tasks */
	Scheduler_SetTaskMode(CDC_Task, TASK_STOP);
	Scheduler_SetTaskMode(USB_USBTask, TASK_STOP);

	/* Indicate USB not ready */
	//LEDS_PORT &= ~(_BV(TXD_LED_PIN));  //txd on
	//LEDS_PORT |= _BV(RXD_LED_PIN);  //rxd off
}

/*---------------------------------------EVENT_HANDLER(USB_ConfigurationChanged)
 * setup 3 endpoints, start USB_USBTask, update status
 *-----------------------------------------------------------------------------*/

EVENT_HANDLER(USB_ConfigurationChanged)
{
	/* Setup CDC Notification, Rx and Tx Endpoints */
	Endpoint_ConfigureEndpoint(CDC_NOTIFICATION_EPNUM, EP_TYPE_INTERRUPT,
		                       ENDPOINT_DIR_IN, CDC_NOTIFICATION_EPSIZE,
	                           ENDPOINT_BANK_SINGLE);

	Endpoint_ConfigureEndpoint(CDC_TX_EPNUM, EP_TYPE_BULK,
		                       ENDPOINT_DIR_IN, CDC_TXRX_EPSIZE,
	                           ENDPOINT_BANK_DOUBLE);

	Endpoint_ConfigureEndpoint(CDC_RX_EPNUM, EP_TYPE_BULK,
		                       ENDPOINT_DIR_OUT, CDC_TXRX_EPSIZE,
	                           ENDPOINT_BANK_DOUBLE);

	/* Indicate USB connected and ready */
	//LEDS_PORT |= _BV(TXD_LED_PIN)|_BV(RXD_LED_PIN);  //both off
	

	/* Start CDC task */
	Scheduler_SetTaskMode(CDC_Task, TASK_RUN);
}

/*--------------------------------------EVENT_HANDLER(USB_UnhandledControlPacket)
 * 
 * This handler deals with requests specific to CDC device class,
 *
 * In particular it handles serial setup and control (DTR/RTS) requests. 
 *
 * note: Request and RequestType are declared by the EVENT_HANDLER macro
 *-----------------------------------------------------------------------------*/

#define SET_CONTROL_LINE_STATE_RTS_MASK 0x0002
#define SET_CONTROL_LINE_STATE_DTR_MASK 0x0001

EVENT_HANDLER(USB_UnhandledControlPacket)
{
	uint8_t* LineCodingData = (uint8_t*)&LineCoding;
    uint16_t wValue;
	
	wValue = Endpoint_Read_Word_LE();
//	Endpoint_Ignore_Word();

	/* Process CDC specific control requests */
	switch (bRequest)
	{
		case GET_LINE_CODING:
			if (bmRequestType == (REQDIR_DEVICETOHOST | REQTYPE_CLASS | REQREC_INTERFACE))
			{
				Endpoint_ClearSetupReceived();

				for (uint8_t i = 0; i < sizeof(LineCoding); i++)
				  Endpoint_Write_Byte(*(LineCodingData++));	
				
				Endpoint_ClearSetupIN();
				while (!(Endpoint_IsSetupINReady()));
				
				while (!(Endpoint_IsSetupOUTReceived()));
				Endpoint_ClearSetupOUT();
			}
			
			break;
		case SET_LINE_CODING:
			if (bmRequestType == (REQDIR_HOSTTODEVICE | REQTYPE_CLASS | REQREC_INTERFACE))
			{
				Endpoint_ClearSetupReceived();

				while (!(Endpoint_IsSetupOUTReceived()));

				for (uint8_t i = 0; i < sizeof(LineCoding); i++)
				  *(LineCodingData++) = Endpoint_Read_Byte();

				Endpoint_ClearSetupOUT();

				ReconfigureUSART();

				Endpoint_ClearSetupIN();
				while (!(Endpoint_IsSetupINReady()));
			}
	
			break;
		case SET_CONTROL_LINE_STATE:
			if (bmRequestType == (REQDIR_HOSTTODEVICE | REQTYPE_CLASS | REQREC_INTERFACE))
			{
				if (wValue & SET_CONTROL_LINE_STATE_DTR_MASK) {
//				     DTR_PIN_PORT &= ~(_BV(DTR_PIN));
					 DTR_Flag = true;
//				     Scheduler_SetTaskMode(CDC_DTR_PULSE_Task, TASK_RUN);
					 //LEDs_TurnOffLEDs(STATUS_RED_LED);
			    } //else {
				//	 Handle Unset DTR request
				//}
				//if (wValue & SET_CONTROL_LINE_STATE_RTS_MASK) {
				//     Handle Set RTS
			    //} else {
				//     Handle UnSet RTS
				//}
				
				Endpoint_ClearSetupReceived();
				
				Endpoint_ClearSetupIN();
				while (!(Endpoint_IsSetupINReady()));
			}
	
			break;
	}
}

/*-----------------------------------------------------------------------------
 **************************** SERIAL HANDLING SECTION **************************
 *-----------------------------------------------------------------------------*/

/*----------------------------------------------------------------TASK(CDC_Task)
 * Moves the bits from USB to serial and back.
 *-----------------------------------------------------------------------------*/

TASK(CDC_Task)
{
	if (USB_IsConnected)
	{
		/* Select the Serial Rx Endpoint */
		Endpoint_SelectEndpoint(CDC_RX_EPNUM);
		
		if (Endpoint_ReadWriteAllowed())
		{
			/* Read the recieved data endpoint into the transmission buffer */
			while (Endpoint_BytesInEndpoint())
			{
				/* Wait until the buffer has space for a new character */
				while (!((BUFF_STATICSIZE - Rx_Buffer.Elements)));
			
				/* Store each character from the endpoint */
				Buffer_StoreElement(&Rx_Buffer, Endpoint_Read_Byte());
				Flash_RXD(); // do the pov thing
			}
			
			/* Clear the endpoint buffer */
			Endpoint_ClearCurrentBank();
		}
		
		/* Check if Rx buffer contains data */
		if (Rx_Buffer.Elements)
		{
			/* Initiate the transmission of the buffer contents if USART idle */
			//if (UCSR1A & (1 << TXC1))
			//  Serial_TxByte(Buffer_GetElement(&Rx_Buffer));

			/* Initiate the transmission of the buffer contents if USART idle */
			if (!(Transmitting))
			{
				Transmitting = true;
				Serial_TxByte(Buffer_GetElement(&Rx_Buffer));
			}
		}

		/* Select the Serial Tx Endpoint */
		Endpoint_SelectEndpoint(CDC_TX_EPNUM);

		/* Check if the Tx buffer contains anything to be sent to the host */
		if (Tx_Buffer.Elements)
		{
			/* Wait until Serial Tx Endpoint Ready for Read/Write */
			while (!(Endpoint_ReadWriteAllowed()));
			
			/* Check before sending the data if the endpoint is completely full */
			bool IsFull = (Endpoint_BytesInEndpoint() == CDC_TXRX_EPSIZE);
			
			/* Write the transmission buffer contents to the recieved data endpoint */
			while (Tx_Buffer.Elements && (Endpoint_BytesInEndpoint() < CDC_TXRX_EPSIZE))
			  Endpoint_Write_Byte(Buffer_GetElement(&Tx_Buffer));
			
			/* Send the data */
			Endpoint_ClearCurrentBank();

			/* If a full endpoint was sent, we need to send an empty packet afterwards to terminate the transfer */
			if (IsFull)
			{
				/* Wait until Serial Tx Endpoint Ready for Read/Write */
				while (!(Endpoint_ReadWriteAllowed()));

				/* Send an empty packet to terminate the transfer */
				Endpoint_ClearCurrentBank();
			}
		}
	}
}

/*------------------------------------------------------------ISR(USART1_TX_vect)
 * stuff the next outgoing byte into UDR
 *-----------------------------------------------------------------------------*/
ISR(USART1_TX_vect)
{
	/* Send next character if avaliable */
	if (Rx_Buffer.Elements)
	  UDR1 = Buffer_GetElement(&Rx_Buffer);
	else
	  Transmitting = false;
}

/*------------------------------------------------------------ISR(USART1_RX_vect)
 * put the byte into the buffer for CDC_Taks to send. 
 *-----------------------------------------------------------------------------*/
ISR(USART1_RX_vect)
{
	/* Character recieved, store it into the buffer */
	Buffer_StoreElement(&Tx_Buffer, UDR1);
	Flash_TXD(); // do the pov thing
}
		
/*-------------------------------------------------------TASK(CDC_DTR_PULSE_Task)
 * Insure that DTR pulse is at least DTR_MILISECONDS long
 *-----------------------------------------------------------------------------*/

TASK(CDC_DTR_PULSE_Task)
{
	static SchedulerDelayCounter_t DelayCounter; 

    if (DTR_Flag) { 
		Scheduler_ResetDelay(&DelayCounter);
		DTR_Flag=false;
	} else if (Scheduler_HasDelayElapsed(DTR_MILISECONDS, &DelayCounter)) {
//		DTR_PIN_PORT |= _BV(DTR_PIN); // PULL DTR HIGH
		Scheduler_ResetDelay(&DelayCounter);
		Scheduler_SetTaskMode(CDC_DTR_PULSE_Task, TASK_STOP);
	}	
}
		
/*----------------------------------------------------void ReconfigureUSART(void)
 * handle the reconfiguration request 
 *-----------------------------------------------------------------------------*/

void ReconfigureUSART(void)
{
	uint8_t ConfigMask = 0;

	/* Determine parity - non odd/even parity mode defaults to no parity */
	if (LineCoding.ParityType == Parity_Odd)
	  ConfigMask = ((1 << UPM11) | (1 << UPM10));
	else if (LineCoding.ParityType == Parity_Even)
	  ConfigMask = (1 << UPM11);

	/* Determine stop bits - 1.5 stop bits is set as 1 stop bit due to hardware limitations */
	if (LineCoding.CharFormat == TwoStopBits)
	  ConfigMask |= (1 << USBS1);

	/* Determine data size - 5, 6, 7, or 8 bits are supported */
	if (LineCoding.DataBits == 6)
	  ConfigMask |= (1 << UCSZ10);
	else if (LineCoding.DataBits == 7)
	  ConfigMask |= (1 << UCSZ11);
	else if (LineCoding.DataBits == 8)
	  ConfigMask |= ((1 << UCSZ11) | (1 << UCSZ10));
	
	/* Enable double speed, gives better error percentages at 8MHz */
	UCSR1A = (1 << U2X1);
	
	/* Enable transmit and receive modules and interrupts */
	UCSR1B = ((1 << TXCIE1) | (1 << RXCIE1) | (1 << TXEN1) | (1 << RXEN1));

	/* Set the USART mode to the mask generated by the Line Coding options */
	UCSR1C = ConfigMask;
	
	/* Set the USART baud rate register to the desired baud rate value */
	UBRR1  = SERIAL_2X_UBBRVAL((uint16_t)LineCoding.BaudRateBPS);
}

/*-----------------------------------------------------------------------------
 **************************** BLINKIN LIGHTS SECTION **************************
 *-----------------------------------------------------------------------------*/

/*-------------------------------------------------------------------Flash_TXD()
 * Turn the TXD on and let the schedular keep in on
 *-----------------------------------------------------------------------------*/

void Flash_TXD(void)
{
	//LEDS_PORT &= ~(_BV(TXD_LED_PIN));  //txd on
	TXD_Flag = true;
	Scheduler_SetTaskMode(CDC_TXD_POV_Task, TASK_RUN); // do the pov thing
}

/*-------------------------------------------------------------------Flash_RXD()
 * Turn the RXD on and let the schedular keep in on
 *-----------------------------------------------------------------------------*/

void Flash_RXD(void)
{
//	LEDS_PORT &= ~(_BV(RXD_LED_PIN));  //rxd on
	RXD_Flag = true;
	Scheduler_SetTaskMode(CDC_RXD_POV_Task, TASK_RUN);
}	

/*---------------------------------------------------------TASK(CDC_TXD_POV_Task)
 * This task makes sure that the TXD light stays on long enough to be visible
 *-----------------------------------------------------------------------------*/
TASK(CDC_TXD_POV_Task)
{
	static SchedulerDelayCounter_t DelayCounter; 
	
    if (TXD_Flag) { 
		Scheduler_ResetDelay(&DelayCounter);
		TXD_Flag=false;
	} else if (Scheduler_HasDelayElapsed(POV_MILISECONDS, &DelayCounter)) {
		
//		LEDS_PORT |= _BV(TXD_LED_PIN);  //txd off
		Scheduler_ResetDelay(&DelayCounter);
		Scheduler_SetTaskMode(CDC_TXD_POV_Task, TASK_STOP);
	}	
}
/*---------------------------------------------------------TASK(CDC_RXD_POV_Task)
 * This task makes sure that the TXD light stays on long enough to be visible
 *-----------------------------------------------------------------------------*/

TASK(CDC_RXD_POV_Task)
{
	static SchedulerDelayCounter_t DelayCounter; 
	
    if (RXD_Flag) { 
		Scheduler_ResetDelay(&DelayCounter);
		RXD_Flag=false;
	} else if (Scheduler_HasDelayElapsed(POV_MILISECONDS, &DelayCounter)) {
//		LEDS_PORT |= _BV(RXD_LED_PIN);  //rxd off
		Scheduler_ResetDelay(&DelayCounter);
		Scheduler_SetTaskMode(CDC_RXD_POV_Task, TASK_STOP);
	}	
}

